
import sys
import os
import glob

from PySide6 import QtCore, QtWidgets, QtGui, QtCharts
from PySide6.QtCore import QPointF, Qt
from PySide6.QtGui import QPainter
from PySide6.QtWidgets import QMainWindow, QApplication
from PySide6.QtCharts import QChart, QChartView, QLineSeries
import pandas as pd


class ChartData:
	def __init__(self):
		self.chart = None
		self.xAxis = None	
		self.yAxis = None	
		self.chartView = None			



class MyWidget(QtWidgets.QMainWindow):
	def __init__(self):
		super().__init__()

		main = QtWidgets.QWidget(self)
		layout = QtWidgets.QVBoxLayout(self)

		btnRefresh = QtWidgets.QPushButton("&Refresh")
		btnRefresh.clicked.connect(self.onRefreshClicked)

		self.statFileTabs = QtWidgets.QTabWidget(self)
		self.lastTab = 0

		layout.addWidget(btnRefresh)		
		layout.addWidget(self.statFileTabs)

		self.charts = { }
		self.tabs = { }

		main.setLayout(layout)	
		self.resize(1000, 800)		
		self.setCentralWidget(main)		
		self.initializeCharts()
		self.refreshStatsData()

	def onRefreshClicked(self):
		self.refreshStatsData()

	def refreshStatsData(self):
		for fn in glob.glob("out/Stats/*.txt"):			
			data0 = pd.read_csv(fn, sep='\t', index_col=0)			
			for col in data0:
				chartKey = fn + ":" + col

				series = QLineSeries()
				for index,value in data0[col].iteritems():
					series.append(index, value)

				cd = self.charts[chartKey]
				cd.chart.removeAllSeries()
				cd.chart.addSeries(series)
				cd.xAxis.setRange(data0[col].index.array.min(), data0[col].index.array.max())	
				cd.yAxis.setRange(data0[col].array.min(), data0[col].array.max())	
				series.attachAxis(cd.xAxis)
				series.attachAxis(cd.yAxis)

	def initializeCharts(self):
		
		for fn in glob.glob("out/Stats/*.txt"):
			scrollArea = QtWidgets.QScrollArea(self)
			tabPage = QtWidgets.QWidget(self)
			layout = QtWidgets.QVBoxLayout(self)
			data0 = pd.read_csv(fn, sep='\t', index_col=0)
			scrollArea.setWidgetResizable(True)

			for col in data0:
				chartKey = fn + ":" + col

				#series = QLineSeries()
				#for index,value in data0[col].iteritems():
				#	series.append(index, value)

				chart = QChart()
				chart.legend().hide()
				chart.setTitle(col)
				
				xAxis = QtCharts.QValueAxis()
				yAxis = QtCharts.QValueAxis()
				
				xAxis.setTitleText(data0.index.name)
				yAxis.setTitleText(col)

				xAxis.setLabelFormat("%g")
				yAxis.setLabelFormat("%g")

				xAxis.setTickCount(5)
				yAxis.setTickCount(5)
				
				chart.addAxis(xAxis, Qt.AlignBottom)
				chart.addAxis(yAxis, Qt.AlignLeft)						

				chartView = QChartView(chart)
				chartView.setRenderHint(QPainter.Antialiasing)				
				chartView.setMinimumSize(100, 300)
				chartView.setMaximumSize(1000000, 300)
				
				cd = ChartData()
				cd.chart = chart
				cd.xAxis = xAxis
				cd.yAxis = yAxis
				cd.chartView = chartView

				self.charts[chartKey] = cd
				
				layout.addWidget(chartView)
			
			tabPage.setLayout(layout)
			scrollArea.setWidget(tabPage)
			self.statFileTabs.addTab(scrollArea, os.path.basename(fn))


def run():
	app = QtWidgets.QApplication([])

	widget = MyWidget()    
	#widget.setWindowState(QtCore.Qt.WindowMaximized)
	widget.show()
	
	sys.exit(app.exec())

if __name__ == "__main__":
	
	run()
	


import os
import re
from typing import Optional
from pydantic import BaseModel
from enum import Enum
import logging
from . import mstarvtk

class DataTimeEnum(str, Enum):
	init = "init"
	slice = "slice"
	volume = "volume"

class DataTypeEnum(str, Enum):
	walls = "walls"
	bc = "bc"
	slice = "slice"
	volume = "volume"
	particles = "particles"
	body = "body"
	unknown = "unknown"

class VtkDataTypeEnum(str, Enum):
	point = "point"
	cell = "cell"

class DataVariable(BaseModel):
	name: str
	type: VtkDataTypeEnum
	number_components: int = 1
	component_index: int = 0
	minimum: Optional[float]
	maximum: Optional[float]

class DataSource(BaseModel):

	filename: str
	name: str
	time_group: DataTimeEnum
	data_type: DataTypeEnum
	variables: list[DataVariable] = []


def GetFieldDataRanges(fd=None):
	r = []
	for i in range(fd.GetNumberOfArrays()):
		a = fd.GetArray(i)
		ncomp = a.GetNumberOfComponents()
		for ci in range(ncomp):
			amin,amax = a.GetRange(ci)
			r.append( (a.GetName(), ncomp, ci, amin, amax ) )
	return r

def GetVariableRanges(dset):
	for name,ncomp,compi,compmin,compmax in GetFieldDataRanges(dset.GetCellData()):	
		yield DataVariable(name=name, type=VtkDataTypeEnum.cell, number_components=ncomp, component_index=compi, minimum=compmin, maximum=compmax)
	for name,ncomp,compi,compmin,compmax in GetFieldDataRanges(dset.GetPointData()):	
		yield DataVariable(name=name, type=VtkDataTypeEnum.point, number_components=ncomp, component_index=compi, minimum=compmin, maximum=compmax)

def ReduceMinMaxVariables(varAgg: list[DataVariable]) -> list[DataVariable]:
	"""
	Reduce an aggregated list of data variables by taking min/max on variable range
	"""

	vars = {}
	for v in varAgg:
		key = (v.type, v.name, v.number_components, v.component_index)

		if key in vars:
			vars[key].minimum = min(vars[key].minimum, v.minimum)
			vars[key].maximum = max(vars[key].maximum, v.maximum)
		
		vars[key] = v
	
	return [v for k,v in vars.items()]	

def GetDataObjectVariableRanges(dobj) -> list[DataVariable]:
	
	items = []
	if dobj.IsA("vtkCompositeDataSet"):

		# reduce variable min/max over all blocks
		it = dobj.NewIterator()
		it.InitTraversal()
		allvars = []
		while not it.IsDoneWithTraversal():
			allvars.extend(GetVariableRanges(it.GetCurrentDataObject()))
			it.GoToNextItem()
		
		items = ReduceMinMaxVariables(allvars)
		
	else:
		for v in GetVariableRanges(dobj):		 
			items.append(v)
			
	return items


def GetPvdDataSource(fn="", last_only=True) -> DataSource:
	"""
	"""

	fullpath = fn
	filename = os.path.basename(fn)
	obj_name = ""
	timegroup = DataTimeEnum.init
	type = DataTypeEnum.bc	
	color_by_variable = False
	solid_color_default = "lightgrey"

	solverTypeToEnum = {
		"particles": DataTypeEnum.particles,
		"movingbody": DataTypeEnum.body,
		"x": DataTypeEnum.slice,
		"y": DataTypeEnum.slice,
		"z": DataTypeEnum.slice,
	}

	if filename == "Walls.stl":
		timegroup = DataTimeEnum.init
		type = DataTypeEnum.walls
		obj_name = "Walls"
	elif filename == "BoundaryConditions.pvd":
		timegroup = DataTimeEnum.init
		type = DataTypeEnum.bc
		obj_name = "Boundary Conditions"
	elif filename == "Volume.pvd":
		timegroup = DataTimeEnum.volume
		type = DataTypeEnum.volume
		color_by_variable = True
		obj_name = "Volume"
	else:
		
		if (m := re.match(r"^(Slice|Volume)([^_]+)_(.*)\.pvd$", filename)) is not None:
									
			try:
				timegrpstr = m.group(1).lower()
				typestr = m.group(2).lower()

				timegroup = DataTimeEnum[timegrpstr]
				type = solverTypeToEnum[typestr]
				obj_name = m.group(3)

			except KeyError:
				logging.warn("Could not determine meta data: %s", fn)
				return None
		
		else:
			logging.warn("Unknown output filename: %s" % fn)
			return None

	pvdr = mstarvtk.PvdReader(fn)
	times = pvdr.get_times()
	if len(times):
		if last_only:
			times = [ times[-1] ]
	else:
		return None
	
	vars = []
	for t in times:
		dataobj = pvdr.get_data(t)
		vars.extend( GetDataObjectVariableRanges(dataobj) )

	vars = ReduceMinMaxVariables(vars)

	return DataSource(filename=fullpath, name=obj_name, time_group=timegroup, data_type=type, variables=vars)
	





from .mstarvtk import PvdReader
import pyarrow as pa
import vtk
from vtk.numpy_interface import dataset_adapter
import polars as po
import numpy as np
import os

parquet_compression = "snappy"

def convert(pvdfn="", single_file="", separate_files_dir="", output_format="", id_select=[]):
	"""
	Convert a particles pvd file into a pyarrow table
	"""

	pvdReader = PvdReader(pvdfn)

	single_obj = None
	do_concat_file = len(single_file) > 0
	do_sep_files = len(separate_files_dir) > 0

	for time in pvdReader.get_times():
		for fn in pvdReader .get_filenames(time):

			reader = vtk.vtkXMLPolyDataReader()
			reader.SetFileName(fn)
			reader.Update()
			
			dataobj = dataset_adapter.WrapDataObject(reader.GetOutput())
			
			datad = {}
			pointname = [ "x", "y", "z" ]
			nrows = 0
			for compIndex in range(0, dataobj.Points.shape[1]):
				nrows = dataobj.Points.shape[0]
				datad[pointname[compIndex]] = pa.array(dataobj.Points[:,compIndex])

			for key in dataobj.PointData.keys():

				dataarr = dataobj.PointData[key]
				if len(dataarr.shape) == 1:
					datad[key] = pa.array(dataarr)
				else:
					for compIndex in range(0, dataarr.shape[1]):
						datad[key + " " + str(compIndex)] = pa.array(dataarr[:,compIndex])
					
			datad["time"] = np.repeat(time, nrows)
			df = po.DataFrame(datad)
			
			if id_select is not None and len(id_select):
				df = df.filter(po.col("ID").is_in(id_select))

			if do_concat_file:
				if single_obj is None:
					single_obj = df
				else:
					single_obj = po.concat([single_obj, df], how="vertical")

			if do_sep_files:
				if output_format.lower() == "csv":
					out_fn = os.path.join(separate_files_dir, "time_%g.csv" % time)
					os.makedirs(os.path.dirname(out_fn), exist_ok=True)
					df.write_csv(out_fn)
				elif output_format.lower() == "parquet":
					out_fn = os.path.join(separate_files_dir, "time_%g.parquet" % time)
					os.makedirs(os.path.dirname(out_fn), exist_ok=True)
					df.write_parquet(out_fn, compression=parquet_compression)


			#print (df.head())

	if single_obj is not None:
		if os.path.splitext(single_file)[1].lower() == ".csv":
			single_obj.write_csv(single_file)
		elif os.path.splitext(single_file)[1].lower() == ".parquet":
			single_obj.write_parquet(single_file, compression=parquet_compression)

def cli():
	
	import argparse
	import glob

	parser = argparse.ArgumentParser()
	parser.add_argument("-i", "--input-dir", default="out", help="where is solver output located?")
	parser.add_argument("-o", "--output-dir", default="out/Processed", help="where output files are written")
	parser.add_argument("--format", default='csv', help="output file format", choices=["csv", "parquet"])
	parser.add_argument("--freq", default='slice', help="time frequency selection", choices=["slice", "volume", "all"])
	parser.add_argument("--separate", default=False, action='store_true', help="store files for each time. Note that a new sub-directory will be created for each particle set")
	parser.add_argument("-id", "--select-id", action='append', help="Select particle by IDs during conversion")
	clargs = parser.parse_args()

	if clargs.select_id is not None and len(clargs.select_id) > 0:
		clargs.select_id = list(map(int, clargs.select_id))

	fileglobs = []

	if clargs.freq in [ "volume", "all"] :
		fileglobs.append("VolumeParticles_*.pvd")
	if clargs.freq in [ "slice", "all"] :
		fileglobs.append("SliceParticles_*.pvd")

	for fglob in fileglobs:
		for fn in glob.glob(os.path.join(clargs.input_dir, "Output", fglob)):

			single_fn = os.path.join(clargs.output_dir, os.path.splitext(os.path.basename(fn))[0])
			separate_dir = ""
			if clargs.separate:
				separate_dir = single_fn
				single_fn = ""		
			else:
				single_fn += "." + clargs.format

			print("-- Processing file:", fn)
			convert(fn, 
							single_file=single_fn, 
							separate_files_dir=separate_dir, 
							output_format=clargs.format, 
							id_select=clargs.select_id)


if __name__ == "__main__":

	convert("out/Output/SliceParticles_Particles.pvd", 
			single_file="out/Processed/particledata.parquet", 
			separate_files_dir="out/Processed/ParticleData/", 
			output_format="parquet")


"""
- Read configuration of post processed output
- Run all plot generation
"""

import argparse
import itertools
import xml.etree.ElementTree as ET
import logging
import vtk
import sys
import os
import glob
import io
from PIL import Image, ImageFont, ImageDraw
import polars as pl
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import subprocess
import json
from pprint import pprint

from . import mstarvtk, vtk_meta
from .config import *
from .vtk_meta import DataTypeEnum, DataTimeEnum, DataVariable

def get_file_path(outdir, fn):	
	if not os.path.isabs(fn) and not os.path.isfile(fn):
		return os.path.join(outdir, fn)
	return fn

def get_pipeline_times(outdir, pl):
	if os.path.splitext(os.path.basename(pl.file))[1].lower() == ".pvd":	
		fn = get_file_path(outdir, pl.file)
		readerPvd = mstarvtk.PvdReader(fn)
		return readerPvd.get_times()
	return []

def get_plot_times(outdir, plotConf):
	all_pipelines = itertools.chain(plotConf.block_pipelines, plotConf.moving_pipelines, plotConf.particle_pipelines, plotConf.volume_pipelines)
	all_times = []
	for pl in all_pipelines:
		all_times.append((pl.file, get_pipeline_times(outdir, pl) ))
	return all_times


def create_example_config():
	conf = BatchPostConfig()

	c1 = VTKPlotConfig(name="Slice Y Plot", 
						variables=[VTKVariable(name="Velocity Magnitude (m/s)", min=0, max=1, discrete=True)],
						block_pipelines=[BlockDataPipeline(file="SliceY_0.167.pvd", color_by=ColorByEnum.variable)],
						standard_views=[ViewStandard(name="Top", eye_side=DirEnum.yp, up=DirEnum.zp)],
						video_options=VideoOption(enabled=True))	

	conf.plots = [ c1 ] 
	return conf


def create_example_volume_config():
	conf = BatchPostConfig()

	c1 = VTKPlotConfig(name="Volume", 
						variables=[VTKVariable(name="Velocity Magnitude (m/s)", min=0, max=1)],
						volume_pipelines=[VolumePipeline(file="Volume.pvd")],
						standard_views=[ViewStandard(name="Top", eye_side=DirEnum.yp, up=DirEnum.zp)],
						video_options=VideoOption(enabled=True))	
						
	conf.plots = [ c1 ] 
	return conf

def create_example_stat_plots_config():

	c1 = VTKPlotConfig(name="Slice Y Plot", 
						variables=[VTKVariable(name="Velocity Magnitude (m/s)", min=0, max=1, discrete=True)],
						block_pipelines=[BlockDataPipeline(file="SliceY_0.167.pvd", color_by=ColorByEnum.variable)],
						standard_views=[ViewStandard(name="Top", eye_side=DirEnum.yp, up=DirEnum.zp)],
						video_options=VideoOption(enabled=True))

	conf = BatchPostConfig(
		auto_stat_plots=True,
		auto_pdf_report=True,
		default_stat_plot_style=StatPlotStyle(
				legend_location=LegendLocEnum.upper_right,
				size=(1000, 300),
				style=["ggplot"]),
		stat_plots=[
		StatPlot(name="Impeller Forces", 
			x_title="Time [s]",
			y_title="Force [N]",
			chart_title="Moving Body Forces",
			series=[
				StatPlotDataSeries(filename="MovingBody_Moving Body.txt", y="Force X [N]"),
				StatPlotDataSeries(filename="MovingBody_Moving Body.txt", y="Force Y [N]"),
				StatPlotDataSeries(filename="MovingBody_Moving Body.txt", y="Force Z [N]")
				]),],
		plots=[VTKPlotConfig(name="Slice Y Plot", time=TimesEnum.last,
						variables=[VTKVariable(name="Velocity Magnitude (m/s)", min=0, max=1, discrete=True)],
						block_pipelines=[BlockDataPipeline(file="SliceY_0.167.pvd", color_by=ColorByEnum.variable)],
						standard_views=[ViewStandard(name="Top", eye_side=DirEnum.yp, up=DirEnum.zp)],
						video_options=VideoOption(enabled=True))]
		)

	return conf

def RgbToFloat(rgb):
	return max(0.0, min(1.0, rgb / 255.0))

def RgbValToFloat(rgb):
	return (RgbToFloat(rgb.red), RgbToFloat(rgb.green), RgbToFloat(rgb.blue))

def CreateOpacityFunction(opacity, variable):	

	if opacity.type == OpacityFuncType.ramp:
		data = [(opacity.start, 0.0), 
				(opacity.end, 1.0) ]
	elif opacity.type == OpacityFuncType.constant:		
		data = [ (0.0, opacity.constant_value), (1.0, opacity.constant_value) ]
	elif opacity.type == OpacityFuncType.smooth:
		xs = np.linspace(0, 1, 128)
		e0 = opacity.start
		e1 = opacity.end
		ys = [ smootherStep(e0, e1, xi) for xi in xs  ]
		data = zip(xs, ys)
	else:
		raise ValueError ("Opacity type not supported: " + str(opacity.type))

	opfunc = vtk.vtkPiecewiseFunction()	
	vrange = variable.max - variable.min
	for x,y in data:
		xval = variable.min + x * vrange
		opfunc.AddPoint(xval, y)

	return opfunc



def CreateAllBatchConfig(case_dir="", do_slice=False, do_volume=False, do_particle=False, do_body=False, exclude_rare=True, range_last_only=True, always_show_particles=False) -> BatchPostConfig:
	"""
	
	:param case_dir: case directory path
	:param do_slice: process slices
	:param do_particle: process particles
	:param do_body: process moving body field data
	:param exclude_rare: exclude rarely used fields and files
	:param range_last_only: determine color scale ranges from last time step only, otherwise, use all times
	"""

	conf = BatchPostConfig()

	conf.auto_pdf_report = True
	conf.auto_stat_plots = True
	conf.default_stat_plot_style = StatPlotStyle(
				legend_location=LegendLocEnum.upper_right,
				size=(1000, 300),
				style=["ggplot"])

	input_dir = os.path.join(case_dir, "out/Output")

	excludedFiles = []
	excludedFields = []
	if exclude_rare:
		excludedFiles = [
			"BoundaryConditions.pvd"
		]

		excludedFields = [
			"vtkGhostType",
			"Volume Fraction (-)",
			"ID",
			"OriginID",
			"Nearest Neighbor Distance (m)"
		]

	srcs: list[vtk_meta.DataSource] = []
	for fn in glob.glob(os.path.join(input_dir, "*.pvd")):

		if os.path.basename(fn) in excludedFiles:
			continue

		if (metad := vtk_meta.GetPvdDataSource(fn, range_last_only)) is not None:
			srcs.append(metad)

	common_views = [ 
		ViewStandard(name="YP", eye_side=DirEnum.yp, up=DirEnum.zp), 
		ViewStandard(name="XP", eye_side=DirEnum.xp, up=DirEnum.yp),
		ViewStandard(name="ZP", eye_side=DirEnum.zp, up=DirEnum.yp)
	]

	if do_slice:

		proto = VTKPlotConfig(name="Slice Plot")
		proto.video_options.enabled = True		

		vars = []
		for src in srcs:
			if src.time_group == DataTimeEnum.slice:
				if src.data_type == DataTypeEnum.walls:
					proto.stl_pipelines.append(StaticStlPipeline(file=src.filename))
				elif src.data_type == DataTypeEnum.slice:
					vars.extend(src.variables)
					proto.block_pipelines.append(BlockDataPipeline(file=src.filename, color_by=ColorByEnum.variable))
				elif src.data_type == DataTypeEnum.body:
					proto.moving_pipelines.append(MovingBodyPipeline(file=src.filename, color_by=ColorByEnum.solid_color))
				elif always_show_particles and src.data_type == DataTypeEnum.particles:
					proto.particle_pipelines.append(ParticlesPipeline(file=src.filename, color_by=ColorByEnum.solid_color))
		
		vars = vtk_meta.ReduceMinMaxVariables(vars)

		for v in vars:
			if v.name in excludedFields:
				continue

			proto.variables.append(VTKVariable(name=v.name, 
												min=v.minimum, 
												max=v.maximum, 
												component=v.component_index, 
												discrete=True, 
												number_components=v.number_components))

		proto.standard_views = common_views

		conf.plots.append(proto)

	if do_volume:

		proto = VTKPlotConfig(name="Volume Plot")
		proto.video_options.enabled = True

		vars = []
		for src in srcs:
			if src.time_group == DataTimeEnum.volume:
				if src.data_type == DataTypeEnum.walls:
					proto.stl_pipelines.append(StaticStlPipeline(file=src.filename))
				elif src.data_type == DataTypeEnum.body:
					proto.moving_pipelines.append(MovingBodyPipeline(file=src.filename, color_by=ColorByEnum.solid_color))
				elif always_show_particles and src.data_type == DataTypeEnum.particles:
					proto.particle_pipelines.append(ParticlesPipeline(file=src.filename, color_by=ColorByEnum.solid_color))
				elif src.data_type == DataTypeEnum.volume:
					vars = src.variables
					proto.volume_pipelines.append(VolumePipeline(file=src.filename))

		vars = vtk_meta.ReduceMinMaxVariables(vars)

		for v in vars:
			if v.name in excludedFields:
				continue
	
			proto.variables.append(VTKVariable(name=v.name, 
												min=v.minimum, 
												max=v.maximum, 
												component=v.component_index, 
												discrete=False, 
												number_components=v.number_components))

		proto.standard_views = common_views

		conf.plots.append(proto)

	if do_body:

		proto = VTKPlotConfig(name="Body Plot")
		proto.video_options.enabled = True

		vars = []
		for src in srcs:
			if src.data_type == DataTypeEnum.walls:
				proto.stl_pipelines.append(StaticStlPipeline(file=src.filename))			
			elif src.data_type == DataTypeEnum.body:
				vars.extend(src.variables)
				proto.moving_pipelines.append(MovingBodyPipeline(file=src.filename, color_by=ColorByEnum.variable))
			elif always_show_particles and src.data_type == DataTypeEnum.particles:
				proto.particle_pipelines.append(ParticlesPipeline(file=src.filename, color_by=ColorByEnum.solid_color))

		for v in vars:			
			proto.variables.append(VTKVariable(name=v.name, min=v.minimum, max=v.maximum, component=v.component_index, discrete=True))

		proto.standard_views = common_views

		conf.plots.append(proto)

	if do_particle:			
		proto = VTKPlotConfig(name="Particle Plot")
		proto.video_options.enabled = True

		vars = []
		for src in srcs:
			if src.data_type == DataTypeEnum.walls:
				proto.stl_pipelines.append(StaticStlPipeline(file=src.filename))			
			elif src.data_type == DataTypeEnum.body:				
				proto.moving_pipelines.append(MovingBodyPipeline(file=src.filename, color_by=ColorByEnum.solid_color))
			elif src.data_type == DataTypeEnum.particles:
				vars.extend(src.variables)
				proto.particle_pipelines.append(ParticlesPipeline(file=src.filename, color_by=ColorByEnum.variable))

		for v in vars:			
			proto.variables.append(VTKVariable(name=v.name, min=v.minimum, max=v.maximum, component=v.component_index, discrete=True))

		proto.standard_views = common_views

		conf.plots.append(proto)

	return conf


def run():

	mpl.use("agg")
	parser = argparse.ArgumentParser()
	subparsers = parser.add_subparsers(dest="action")

	# JSON based definition of post processing
	fileParser = subparsers.add_parser("file")
	fileParser.add_argument("filename", default="post.json", help="use an input file to declare batch post processing output")
	
	# CLI
	autoParser = subparsers.add_parser("auto")
	autoParser.add_argument("-d", "--directory", default=".", help="case directory")	
	autoParser.add_argument("--all", action='store_true', help="process all data sources (slice,volume,particle,body)")
	autoParser.add_argument("--slice", action='store_true', help="process all slices")
	autoParser.add_argument("--volume", action='store_true', help="process the volume")
	autoParser.add_argument("--particle", action='store_true', help="process all particles")
	autoParser.add_argument("--body", action='store_true', help="process all bodies")
	autoParser.add_argument("--always-show-particles", action='store_true', help="show particles on all plots")
	autoParser.add_argument("--debug", action='store_true', help="iterate all variables in all input files, by default, less commonly used fields exluded from processing")

	# Testing
	testParser = subparsers.add_parser("test")
	testParser.add_argument("--test", action='store_true', help="reserved for internal testing")


	clargs = parser.parse_args()

	case_dir = os.getcwd()	
	conf = None
	if clargs.action == "test":
		conf = create_example_stat_plots_config()

		with open("post.json", 'w') as f:
			json.dump(conf.dict(), f, indent=4, sort_keys=True)	

		print(conf.dict())
	
	elif clargs.action == "auto":
		case_dir = os.path.abspath(clargs.directory)

		if clargs.all:
			clargs.slice = True
			clargs.volume = True
			clargs.particle = True
			clargs.body = True

		conf = CreateAllBatchConfig(clargs.directory, 
									do_slice=clargs.slice, 
									do_volume=clargs.volume, 
									do_particle=clargs.particle, 
									do_body=clargs.body, 
									always_show_particles=clargs.always_show_particles)

	elif clargs.action == "file":
		case_dir = os.path.dirname(os.path.abspath(clargs.filename))
		conf = BatchPostConfig.parse_file(clargs.filename)

	else:

		parser.print_usage()
		sys.exit(1)

	ffmpeg_exe = "ffmpeg"

	ffmpegSearchPaths = [
		os.path.join(os.path.dirname(__file__), "../../ffmpeg.exe")
	]

	for pth in ffmpegSearchPaths:
		if os.path.isfile(pth):
			ffmpeg_exe = pth
			break

	
	input_dir = os.path.join(case_dir, "out")
	output_dir = os.path.join(input_dir, "Processed")

	os.makedirs(output_dir, exist_ok=True)

	logging.getLogger().setLevel(logging.INFO)
	logging.debug(clargs)

	results = ResultsProcessed()

	bounds = None
	try:
		inputxml = ET.parse(os.path.join(case_dir, "input.xml")).getroot()
		lx = float(inputxml.find("./system/domainX").text)
		ly = float(inputxml.find("./system/domainY").text)
		lz = float(inputxml.find("./system/domainZ").text)
		minDomain = tuple(map(float, inputxml.find("./system/domainOrigin").text.split()))
		bounds = (minDomain[0], minDomain[0] + lx, minDomain[1], minDomain[1] + ly, minDomain[2], minDomain[2] + lz) 
		logging.info("Found domain bounds: %s", str(bounds))
	
	except:
		logging.warn("Failed to parse input xml")
		raise

	# dump input post config to output dir
	with open(os.path.join(output_dir, "post.json"), 'w') as f:
		json.dump(conf.dict(), f, indent=4, sort_keys=True)	

	if (default_style := conf.default_stat_plot_style) is None:
		default_style = StatPlotStyle()

	if conf.stat_plots is not None:
		stat_data = { }
		px = 1/plt.rcParams['figure.dpi']  # pixel in inches

		for statPlot in conf.stat_plots:
			for s in statPlot.series:
				fn = s.filename
				if fn not in stat_data:
					final_fn = ""
					if os.path.isabs(fn):
						final_fn = fn
					elif os.path.isfile(final_fn := fn):
						pass
					elif os.path.isfile(final_fn := os.path.join(input_dir, "Stats/", fn)):
						pass							

					if os.path.isfile(final_fn):
						logging.info("Loading file: %s", fn)
						stat_data[fn] = pl.read_csv(final_fn, sep='\t')
					else:
						logging.error("Could not find file: %s", fn)
						raise ValueError()	

		for statPlot in conf.stat_plots:

			statPlotStyle = default_style.override(statPlot.style)

			mpl.style.use(statPlotStyle.style)

			fig, ax = plt.subplots(figsize=(statPlotStyle.size[0]*px,statPlotStyle.size[1]*px))

			xnames = set()
			ynames = set()
			for series in statPlot.series:
				dataframe = stat_data[series.filename]
				if (xcolname := series.x) is None:
					if series.x_column is None:
						raise ValueError("Must specify column name or index")
					else:
						xcolname = dataframe.columns[series.x_column]

				if (ycolname := series.y) is None:
					if series.y_column is None:
						raise ValueError("Must specify column name or index")
					else:
						ycolname = dataframe.columns[series.y_column]

				kwargs = dict()
				if series.line_color is not None:
					kwargs["color"] = series.line_color.ToFloatValue()
				if series.line_width is not None:
					kwargs["linewidth"] = series.line_width
				if series.line_style is not None:
					kwargs["linestyle"] = series.line_style

				xnames.add(xcolname)
				ynames.add(ycolname)
				#print(dataframe.head())
				ax.plot(dataframe.select(xcolname).to_numpy(), dataframe.select(ycolname).to_numpy(), **kwargs, label=ycolname)

			if statPlotStyle.legend_location is not None:
				fig.legend(loc=statPlotStyle.legend_location.value)
			
			if (xtitle := statPlot.x_title) is None:
				if len(xnames) == 1:
					xtitle = list(xnames)[0]
				else:
					logging.warn("Could not determine unique xtitle")

			if (ytitle := statPlot.y_title) is None:
				if len(ynames) == 1:
					ytitle = list(ynames)[0]
				else:
					logging.warn("Could not determine unique ytitle")

			if xtitle is not None:
				ax.set_xlabel(xtitle)

			if ytitle is not None:
				ax.set_ylabel(ytitle)

			if statPlot.chart_title is not None:
				ax.set_title(statPlot.chart_title)

			
			fig.tight_layout()

			img_out_dir = os.path.join(output_dir, "CustomStatPlots")
			outfn = os.path.join(img_out_dir, statPlot.name + ".png")
			os.makedirs(img_out_dir, exist_ok=True)
			fig.savefig(outfn)		

			results.custom_stat_plots.append(ResultStatImage(name=statPlot.name, config=statPlot.copy(), image_filename=outfn))

			plt.close(fig)
			del ax
			del fig

		del stat_data

	if conf.auto_stat_plots:
		stats_dir = os.path.join(input_dir, "Stats/")
		statPlotStyle = default_style
		for fn in glob.glob(os.path.join(stats_dir, "*.txt")):
			df = pl.read_csv(fn, sep='\t')
			xcolname = df.columns[0]
			for i in range(1, len(df.columns)):

				if statPlotStyle.style is not None and len(statPlotStyle.style) > 0:
					mpl.style.use(statPlotStyle.style)

				ycolname = df.columns[i]

				kwag = {}
				if statPlotStyle.size is not None:
					kwag["figsize"] = (statPlotStyle.size[0]*px,statPlotStyle.size[1]*px)

				fig, ax = plt.subplots(**kwag)
				ax.plot(df.select(xcolname).to_numpy(), df.select(ycolname).to_numpy(), label=ycolname)												
				ax.set_xlabel(xcolname)			
				ax.set_title(ycolname)
				
				fig.tight_layout()

				img_out_dir = os.path.join(output_dir, "AutoStatPlots", os.path.splitext(os.path.basename(fn))[0])
				img_fn = mstarvtk.NiceVariableName( ycolname + ".png" )
				outfn = os.path.join(img_out_dir, img_fn)
				os.makedirs(img_out_dir, exist_ok=True)
				fig.savefig(outfn)

				results.auto_stat_plots.append(ResultAutoStatImage(stat_filename=fn, name=ycolname, image_filename=outfn))

				plt.close(fig)

				del ax
				del fig

	if conf.plots is not None:
		for p in conf.plots:			

			outdir = os.path.join(input_dir, "Output")

			title = p.name		
			img_size = (p.size_x, p.size_y)

			colors = vtk.vtkNamedColors()

			views = dict()

			for view in p.custom_views:
				cam = vtk.vtkCamera()
				cam.SetParallelProjection(view.parallel_projection)	
				#cam.SetParallelScale(height)		
				cam.SetViewUp(view.up)
				cam.SetPosition(view.eye_position)
				cam.SetFocalPoint(view.look_at)
				views[view.name] = cam

			for view in p.standard_views:
				m = dict(x=0, y=1, z=2)
				eyeindex = view.eye_side.direction()
				eyeflip = view.eye_side.sign()
				up = view.up.direction()
				cam = mstarvtk.create_camera_tight(eyeindex, eyeflip, up, bounds)
				views[view.name] = cam

			for variable in p.variables:
				
				colorTranFunc = mstarvtk.CreateColorTransferFunction(variable.name, variable.min, variable.max, variable.color_scale, 256, variable.discrete, variable.discrete_n)

				pilFontSize = 18
				scalar_bar_img_2 = mstarvtk.draw_color_bar(colorTranFunc, opacityFunc=None, size=(76, img_size[1]), min=variable.min, max=variable.max, fontsize=pilFontSize)

				# Render the frames directly into a video file
				img_index = 0
				
				data_times_pairs = get_plot_times(outdir, p)
				data_times = []
				for pvdname,times in data_times_pairs:
					data_times = times
					break

				plot_times = []

				if p.time == TimesEnum.all:
					plot_times = data_times
				elif  p.time == TimesEnum.last:	
					plot_times = [ data_times[-1] ]		
				else:
					raise ValueError("Time selection option not supported: " + p.time_option)

				# dictionary of tuples
				# key: view name
				# value: (filename pattern)
				videos = dict()

				video_rate = 1.0

				if len(plot_times) >= 2:
					video_rate = 1.0 / (plot_times[1] - plot_times[0])
				
				video_rate *= p.video_options.speed

				for t in plot_times:

					print("Processing: %s - %s - t=%g" % (p.name, variable.name, t) )

					camera = vtk.vtkCamera()
					camera.SetPosition(1, 1, 1)
					camera.SetFocalPoint(0, 0, 0)

					renderer = vtk.vtkRenderer()
					text_annotation = "  ".join( ( title, variable.name, "t = {0}".format(t) ) )			
					renderer.SetBackground(RgbValToFloat(p.background_color))
					renderer.SetActiveCamera(camera)	
					
					for pipe in p.block_pipelines:
						readerFn = get_file_path(outdir, pipe.file)
						o = mstarvtk.BlockDataPipeline(readerFn, t, False, colorTranFunc)
						o.mapper.SelectColorArray(variable.name)			
						renderer.AddActor(o.actor)
					
					for pipe in p.moving_pipelines:
						readerFn = get_file_path(outdir, pipe.file)
						o = mstarvtk.MovingBodyPipeline(readerFn, t)				
						renderer.AddActor(o.actor)
					
					for pipe in p.stl_pipelines:			
						readerFn = get_file_path(outdir, pipe.file)
						o = mstarvtk.StaticStlPipeline(readerFn)
						renderer.AddActor(o.actor)

					
					for pipe in p.particle_pipelines:			
						readerFn = get_file_path(outdir, pipe.file)
						o = mstarvtk.ParticlesPipeline(readerFn, t)
						renderer.AddActor(o.actor)

					for pipe in p.volume_pipelines:			
						readerFn = get_file_path(outdir, pipe.file)
						opacityFunc = CreateOpacityFunction(pipe.opacity_function, variable)
						o = mstarvtk.VolumeRenderingPipeline(readerFn, t, colorTranFunc, opacityFunc)
						o.mapper.SetScalarModeToUsePointFieldData()
						o.mapper.SelectScalarArray(variable.name)
						renderer.AddActor(o.actor)											
						
					renderWindow = vtk.vtkRenderWindow()
					renderWindow.AddRenderer(renderer)
					renderWindow.SetOffScreenRendering(1)
					renderWindow.SetSize( int(img_size[0]/2), int(img_size[1]/2) ) # why /2 ? 
					renderWindow.SetMultiSamples(4) # what do?

					# An interactor
					renderWindowInteractor = vtk.vtkRenderWindowInteractor()
					renderWindowInteractor.SetRenderWindow(renderWindow)

					# lower left axis orientation
					axes = vtk.vtkAxesActor()
					widget = vtk.vtkOrientationMarkerWidget()
					rgba = [0] * 4
					colors.GetColor('Carrot', rgba)
					widget.SetOutlineColor(rgba[0], rgba[1], rgba[2])
					widget.SetOrientationMarker(axes)
					widget.SetInteractor(renderWindowInteractor)
					widget.SetViewport(0.0, 0.0, 0.1, 0.1)
					widget.SetEnabled(1)

					if p.image_output:

						for viewname,viewcam in views.items():

							renderer.SetActiveCamera(viewcam)
							renderWindow.Render()					

							windowto_image_filter = vtk.vtkWindowToImageFilter()
							windowto_image_filter.SetInput(renderWindow)
							windowto_image_filter.SetScale(2)

							writer = vtk.vtkPNGWriter()
							writer.SetInputConnection(windowto_image_filter.GetOutputPort())
							writer.WriteToMemoryOn()
							writer.Write()

							buf = io.BytesIO()
							buf.write(writer.GetResult())
							buf.seek(0)

							# Read data into PIL image buffer
							vtk_img = Image.open(buf)
							
							w = vtk_img.size[0]
							h = vtk_img.size[1]
							
							total_w = w + scalar_bar_img_2.size[0]
							total_h = h + 20

							# Stack VTK image with scalar bar and annotation in a new image		
							com_img = Image.new("RGBA", size=(total_w, total_h), color=(255,255,255))
							com_img.paste(vtk_img, (0,0) )
							com_img.alpha_composite(scalar_bar_img_2.copy(), dest=(w,0))

							for mark in p.watermarks:
								
								imgfn = mark.image
								if os.path.isfile(imgfn):
									
									watermarkImg = Image.open(imgfn).convert("RGBA")
									marksize = watermarkImg.size							

									desty = 0
									destx = 0
									if mark.location[0] == "t":
										desty = 0
									elif mark.location[0] == "m":
										desty = total_h * 0.5 - 0.5 * marksize[1]
									elif mark.location[0] == "b":
										desty = total_h - marksize[1]

									if mark.location[1] == "l":
										destx = 0
									elif mark.location[1] == "m":
										destx = total_w * 0.5 - 0.5 * marksize[0]
									elif mark.location[1] == "r":
										destx = total_w - marksize[0]
									
									blanklayer = Image.new('RGBA', com_img.size, (0, 0, 0, 0))
									marklayer = blanklayer.copy()
									marklayer.paste(watermarkImg, (int(destx), int(desty)))
									watermark2 = Image.blend(blanklayer, marklayer, mark.alpha)							
									com_img.alpha_composite(watermark2)

									
							font = ImageFont.truetype("arial.ttf", pilFontSize)
							dr = ImageDraw.Draw(com_img)					
							
							dr.text( (0,h), title , anchor='la', font=font, fill="black" )
							dr.text( (total_w/2,h), variable.name , anchor='ma', font=font, fill="black" )
							dr.text( (total_w,h), "t = {0}".format(t) , anchor='ra', font=font , fill="black")
							
							var_name,varunits = mstarvtk.SplitVtkVarName(variable.name)
							varCompDisp = variable.get_component_name()
							fn_prefix = p.name + "_" + var_name
							if len(varCompDisp) > 0:
								fn_prefix += "_" + varCompDisp
							fn_prefix += "_" + viewname

							video_img_fn_fmt = fn_prefix + "_%0" + str(len(str(len(plot_times)))) + "d.png"
							video_img_fn_fmt = video_img_fn_fmt.format(name=var_name, view_name=viewname)

							imageFramesOutDir = os.path.join(output_dir, fn_prefix)

							fn = video_img_fn_fmt % img_index
							fn = os.path.join(imageFramesOutDir, fn)		

							videos[viewname] = ( video_img_fn_fmt, fn_prefix, imageFramesOutDir )

							os.makedirs(imageFramesOutDir, exist_ok=True)

							logging.debug("Saving view: %s", fn)
							com_img.save(fn)

							results.vtk_images.append(ResultVtkImage(time=t, image_filename=fn, config=p, view=viewname))

						img_index += 1
					
				if p.video_options.enabled:

					for viewname,videoTup in videos.items():

						img_pattern, fn_prefix, out_dir = videoTup
						outmp4 = os.path.join(output_dir, fn_prefix + ".mp4")
						cmd = [ffmpeg_exe, 
										"-r", str(video_rate), 
										"-i", img_pattern, 
										"-vcodec", "libx264", 
										"-crf", str(p.video_options.quality),
										"-pix_fmt", "yuv420p", 
										"-y",
										outmp4]

						logging.debug("Running command: %s", " ".join(cmd))
						subprocess.run(cmd, cwd=out_dir)


	if conf.auto_pdf_report:
		from mstarpypost.reports import create_auto_report
		reportConf = AutoReportConfig(title="M-Star CFD Report", results=results)

		modelxmlfn = os.path.join(case_dir, "report.xml")
		if os.path.isfile(modelxmlfn):
			reportConf.model_tree_filename = modelxmlfn
					
		create_auto_report(reportConf, outputfn=os.path.join(output_dir, "Report.pdf"))


	with open(os.path.join(output_dir, "results.json"), 'w') as f:
		json.dump(results.dict(), f, indent=4, sort_keys=True)

if __name__ == "__main__":

	run()



import vtk
from vtk.util.numpy_support import vtk_to_numpy
import xml.etree.ElementTree as ET
import os
import re
import io
import glob
import numpy as np
from matplotlib import cm
import matplotlib as mpl
from collections import namedtuple
import pandas as pd
from PIL import Image, ImageFont, ImageDraw
import math
import logging

def vtk2pil(renderer, width, height):
	
	renderWindow = vtk.vtkRenderWindow()
	renderWindow.AddRenderer(renderer)
	renderWindow.SetOffScreenRendering(1)
	renderWindow.SetSize(width, height)
	renderWindow.SetMultiSamples(4)
	renderWindow.Render()

	windowto_image_filter = vtk.vtkWindowToImageFilter()
	windowto_image_filter.SetInput(renderWindow)    

	writer = vtk.vtkPNGWriter()
	writer.SetInputConnection(windowto_image_filter.GetOutputPort())

	writer.WriteToMemoryOn()
	writer.Write()

	buf = io.BytesIO()
	buf.write(writer.GetResult())
	buf.seek(0)
	
	return Image.open(buf)    

def fig2pil(fig):
    # Save a matplotlib figure to an PIL Image
    import io
    buf = io.BytesIO()
    fig.savefig(buf, format="png")
    buf.seek(0)
    return Image.open(buf)

def make_title_image(title="", subtitle="", size=(500,500)):
    # Make title frame image

    img = Image.new("RGB", size)

    h1 = ImageFont.truetype("arial.ttf", 26)
    h2 = ImageFont.truetype("arial.ttf", 18)
    normal = ImageFont.truetype("arial.ttf", 16)
    draw = ImageDraw.Draw(img)

    if len(title):
        draw.text((size[0]/2, size[1]/2), title, anchor="md", font=h1)

    if len(subtitle):
        draw.text((size[0]/2, size[1]/2), subtitle, anchor="ma", font=h2)

    draw.text((size[0], 0), datetime.now().isoformat(timespec="minutes"), anchor="ra", font=normal)
    draw.text(size, "Created with M-Star CFD", anchor="rd", font=normal)

    return img

class UniqueFloatSet:
	def __init__(self, tolerance=1e-3):
		self.tolerance = tolerance
		self.myset = []
	
	def insert(self, val):
		addVal = False
		for v in self.myset:
			if abs(v - val) > self.tolerance:
				addVal = True
		self.myset.append(val)
		return True
	
	def values(self):
		return self.myset

class ScalarModeEnum:
	Default = 0
	PointData = 1
	CellData = 2
	PointFieldData = 3
	CellFieldData = 4
	FieldData = 5

def AddImageToDims(imgData, bbox, uniqFloats):
	bnds = imgData.GetBounds()
	extent = imgData.GetExtent()
	nx = extent[1] - extent[0]
	ny = extent[3] - extent[2]
	nz = extent[5] - extent[4]
	lx = bnds[1] - bnds[0]
	ly = bnds[3] - bnds[2]
	lz = bnds[5] - bnds[4]
	dx = lx / nx if nx > 0 else 0
	dy = ly / ny if ny > 0 else 0
	dz = lz / nz if nz > 0 else 0
	if dx > 0:
		uniqFloats.insert(dx)
	if dy > 0:
		uniqFloats.insert(dy)
	if dz > 0:
		uniqFloats.insert(dz)
	bbox.AddBounds(bnds)

def GetNiceSamplingBounds(dobj):
	# interrogate a data object and determine good resampling bounds
	
	overallBbox = vtk.vtkBoundingBox()
	uniqueDxs = UniqueFloatSet()
	
	it = dobj.NewIterator()
	it.InitTraversal()
	while not it.IsDoneWithTraversal():
		AddImageToDims(it.GetCurrentDataObject(), overallBbox, uniqueDxs)
		it.GoToNextItem()
	
	maxDx = max(uniqueDxs.values())
	nx = overallBbox.GetLength(0) / maxDx
	ny = overallBbox.GetLength(1) / maxDx
	nz = overallBbox.GetLength(2) / maxDx
	nx = int( max(1, nx) )
	ny = int( max(1, ny) )
	nz = int( max(1, nz) )
	
	return nx, ny, nz
		
def GetFieldDataRanges(fd=None):
	r = []
	for i in range(fd.GetNumberOfArrays()):
		a = fd.GetArray(i)
		ncomp = a.GetNumberOfComponents()
		for ci in range(ncomp):
			amin,amax = a.GetRange(ci)
			r.append( (a.GetName(), ncomp, amin, amax ) )
	return r

def GetVariableRanges(dset):	
	for name,ncomp,compmin,compmax in GetFieldDataRanges(dset.GetCellData()):	
		yield ("CELL", name, ncomp, compmin, compmax )
	for name,ncomp,compmin,compmax in GetFieldDataRanges(dset.GetPointData()):	
		yield ("POINT", name, ncomp, compmin, compmax )

from collections import namedtuple
VariableKey = namedtuple("VariableKey", ["name", "component", "type"])

def GetDataObjectVariableRanges(dobj):
	vars = {}
	if dobj.IsA("vtkCompositeDataSet"):
		it = dobj.NewIterator()
		it.InitTraversal()
		while not it.IsDoneWithTraversal():
			comptype,compname,compi,compmin,compmax = GetVariableRanges(it.GetCurrentDataObject())
			key = VariableKey(name=compname, component=compi, type=comptype)

			if key in vars:
				compmin = min(vars[key][0], compmin)
				compmax = max(vars[key][1], compmax)
			
			vars[key] = (compmin, compmax)
			it.GoToNextItem()
	else:
		for comptype,compname,compi,compmin,compmax in GetVariableRanges(dobj):		 
			key = VariableKey(name=compname, component=compi, type=comptype)
			vars[key] = (compmin, compmax)
	
	for k,v in vars.items():
		yield (k.type, k.name, k.component, v[0], v[1])	

def CellToPointPipeline(inputDataObj):
	if inputDataObj.GetClassName() == "vtkMultiBlockDataSet":
		logging.debug ("-- Multi block data set")
		geoFilter = vtk.vtkCompositeDataGeometryFilter()
		geoFilter.SetInputDataObject(inputDataObj)		
		
		nx, ny, nz = GetNiceSamplingBounds(inputDataObj)			
		resample = vtk.vtkResampleToImage()
		resample.UseInputBoundsOn()
		resample.SetSamplingDimensions(nx, ny, nz)
		resample.SetInputConnection(geoFilter.GetOutputPort())		
		return resample

	else:
		logging.debug ("-- Single block, using vtkCellDataToPointData")
		c2p = vtk.vtkCellDataToPointData()
		c2p.SetInputDataObject(inputDataObj)
		return c2p


def GetAvailableVariables(dset):
	vars = []
	for i in range(dset.GetCellData().GetNumberOfArrays()):
		a = dset.GetCellData().GetAbstractArray(i)
		vars.append( ("CELL", a.GetName(), a.GetNumberOfComponents(), a.GetRange()) )
		
	for i in range(dset.GetPointData().GetNumberOfArrays()):
		a = dset.GetPointData().GetAbstractArray(i)
		vars.append( ("POINT", a.GetName(), a.GetNumberOfComponents(), a.GetRange() ) )
		
	return vars
	
def GetPvdFileMetaData(fn="", last_only=True):
	"""
	Collects meta data from PVD data files

	:param fn: PVD Filename
	:param last_only: If True, only looks at last available time in data, otherwise will read all times
	"""

	pvdr = PvdReader(fn)
	times = pvdr.get_times()
	if len(times):
		if last_only:
			times = [ times[-1] ]
	else:
		return None
	
	alldata = {}
	for t in times:
		data = pvdr.get_data(t)

		alldata[t] = list(GetDataObjectVariableRanges(data))

	return alldata




def PrintAvailableVariables(dset):
	v = GetAvailableVariables(dset)
	print(v)

def NiceVariableName(vname=""):
	bad = ['/', '\\', ':', ' ']
	for b in bad:
		vname = vname.replace(b, "")
	return vname

def SplitVtkVarName(vname=""):
	"""
	Splits up a variable name into (name, unitstr) tuple
	If fails to parse name, just removes bad characters
	"""

	if ( m := re.match(r'^([^(]+)(\s\(([^)]+)\))?$', vname)) is not None:
		if len(m.groups()) == 3:
			return m.group(1),m.group(3)
		elif len(m.groups()) == 3:
			return m.group(1)
	
	return NiceVariableName(vname),""
	


def CreateColorTransferFunction(varname="", minlimit=0, maxlimit=1, colorscale="viridis", ncolors = 256, discrete=False, discreteN=12):	

	ctf = None
	if discrete:
		ctf = vtk.vtkDiscretizableColorTransferFunction ()
		ctf.SetDiscretize(True)
		ctf.SetNumberOfValues(discreteN)
	else:
		ctf = vtk.vtkColorTransferFunction ()

	sampled = np.linspace(minlimit, maxlimit, ncolors)
	norm = mpl.colors.Normalize(minlimit, maxlimit)		
	c = cm.get_cmap(colorscale, ncolors)
	for i in range(ncolors):			
		r,g,b,a = c(norm(sampled[i]))
		ctf.AddRGBPoint(sampled[i], r, g, b)
	return ctf

class PvdReader:
	def __init__(self, pvdfn=""):
		if os.path.isfile(pvdfn):
			self.pvd = pvdfn
		else:
			rel_file = os.path.join("out", "Output", pvdfn)
			if os.path.isfile(rel_file):
				self.pvd = os.path.join("out", "Output", pvdfn)
			else:
				raise ValueError("File does not exist: %s" % pvdfn)
		
	def get_times(self):
		r = ET.parse(self.pvd).getroot()
		timeset = set()
		for d in r.findall(".//DataSet"):
			timeset.add(float(d.get("timestep")))
		times = sorted(timeset)
		return times
		
	def get_filenames(self, time):
		r = ET.parse(self.pvd).getroot()
		
		pvddir = os.path.dirname(self.pvd)
		fns = []
		for d in r.findall(".//DataSet"):
			if time == float(d.get("timestep")):
				fns.append(os.path.join(pvddir, d.get("file")))
			
		return fns
		
	def get_data(self, time):
		
		blockFileNames = self.get_filenames(time)
		blocks =  [] 		
		for fn in blockFileNames:
			extension = os.path.splitext(fn)[1]
			if extension == ".vti":
				reader = vtk.vtkXMLImageDataReader()
			elif extension == ".vtp":
				reader = vtk.vtkXMLPolyDataReader()
			reader.SetFileName(fn)
			reader.Update()
			blocks.append(reader.GetOutput())

		if len(blocks) == 1:
			return blocks[0]
		elif len(blocks) > 1:
			index = 1
			root = vtk.vtkMultiBlockDataSet()		
			for b in blocks:
				root.SetBlock(index, b)
				index += 1
			return root
			
		return None
	
		
	def print_summary(self):
		print("PVD: ", self.pvd)
		for t in self.get_times():
			print("Time: ", t)
			for fn in self.get_filenames(t):
				print("\t",fn)

class MStarFile:
	def __init__(self, fn=""):
		self.filename = fn
		self.name = os.path.basename(fn)
		self.timegroup = ""
		self.type = ""		
		self.color_by_variable = False
		self.solid_color_default = "lightgrey"

		if self.name == "Walls.stl":
			self.timegroup = "Static"
			self.type = "Walls"
		elif self.name == "BoundaryConditions.pvd":
			self.timegroup = "Static"
			self.type = "BC"
		elif self.name == "Volume.pvd":
			self.timegroup = "Volume"
			self.type = "Volume"
			self.color_by_variable = True
		else:
			m = re.match(r"^(Slice|Volume)([^_]+)_(.*)\.pvd$", self.name)
			if m:
				
				self.timegroup = m.group(1)				
				self.type = m.group(2)
				self.name = m.group(3)

			if self.type in[ "X", "Y", "Z" ]:
				self.color_by_variable = True		

	def __str__(self):
		return "Filename: {0}, Time Group: {1}, Type: {2}".format(self.filename, self.timegroup, self.timegroup)

	def data_ranges(self, time):
		if self.filename.endswith(".pvd"):
			reader = PvdReader(self.filename)
			d = reader.get_data(time)
			return GetDataObjectVariableRanges(d)
		return None

	def get_pipeline(self, model):
		o = None
		if self.type in [ "X", "Y", "Z" ]:
			o = BlockDataPipeline(self.filename, model.get_time())
			o.mapper.SelectColorArray(model.variable)			
			#renderer.AddActor(o.actor)

		elif self.type == "MovingBody":
			o = MovingBodyPipeline(self.filename, model.get_time())
			#renderer.AddActor(o.actor)

		elif self.type == "Walls":
			o = StaticStlPipeline(self.filename)
			#renderer.AddActor(o.actor)

		elif self.type == "Particles":
			o = ParticlesPipeline(self.filename, model.get_time())
			#renderer.AddActor(o.actor)

		elif self.type == "Volume":
			o = VolumeRenderingPipeline(self.filename, model.get_time(), model.get_color(), model.get_opacity())
			o.mapper.SetScalarModeToUsePointFieldData()
			o.mapper.SelectScalarArray(model.variable)
			#renderer.AddVolume(o.actor)
		
		return o

	def add_pipeline(self, model, renderer):
		p = self.get_pipeline(model)
		if p is not None:
			renderer.AddActor(p.actor)

class VolumeRenderingPipeline:
	def __init__(self, fn="", time=0.0, color=None, opacity=None):		
		reader = PvdReader(fn)
		pvd_data = reader.get_data(time)
				
		if pvd_data.GetClassName() == "vtkMultiBlockDataSet":
			logging.debug ("-- Multi block data set")
			geoFilter = vtk.vtkCompositeDataGeometryFilter()
			geoFilter.SetInputDataObject(pvd_data)
			geoFilter.Update()
			last = geoFilter
			
			nx, ny, nz = GetNiceSamplingBounds(pvd_data)			
			resample = vtk.vtkResampleToImage()
			resample.UseInputBoundsOn()
			resample.SetSamplingDimensions(nx, ny, nz)
			resample.SetInputConnection(last.GetOutputPort())
			resample.Update()
			last = resample

		else:
			logging.debug ("-- Single block, using vtkCellDataToPointData")
			c2p = vtk.vtkCellDataToPointData()
			c2p.SetInputDataObject(pvd_data)
			c2p.Update()
			last = c2p

		opacityUnitDistance = 0.001
		im = last.GetOutput()
		imDims = im.GetDimensions()
		maxImDims = max(*imDims)
		bbox = vtk.vtkBoundingBox(im.GetBounds())
		if bbox.IsValid() and maxImDims > 0:
			opacityUnitDistance = 10 * bbox.GetDiagonalLength() / float(maxImDims)		

		self.last_filter = last
		mapper = vtk.vtkSmartVolumeMapper()
		mapper.SetInputConnection(last.GetOutputPort())
		mapper.SetBlendModeToComposite()
		mapper.SetRequestedRenderModeToDefault()		

		prop = vtk.vtkVolumeProperty()
		prop.SetColor(color)
		prop.SetScalarOpacity(opacity)
		prop.ShadeOff()
		prop.SetScalarOpacityUnitDistance(0.01)
		prop.SetInterpolationTypeToLinear()

		actor = vtk.vtkVolume()
		actor.SetMapper(mapper)
		actor.SetProperty(prop)

		self.ranges =  [ r for r in GetDataObjectVariableRanges(last.GetOutputDataObject(0)) ]
		self.reader = reader
		self.mapper = mapper
		self.actor = actor	

		logging.debug (str(self.ranges))

	def get_histogram(self):		

		logging.debug ("Get histogram")
		assignAttr = vtk.vtkAssignAttribute()
		assignAttr.SetInputConnection(self.last_filter.GetOutputPort())
		assignAttr.Assign("Volume Fraction (-)", 0, 0) # assign point field to scalar attribute

		logging.debug ("image stencil")
		stencil = vtk.vtkImageToImageStencil()
		stencil.SetInputConnection(assignAttr.GetOutputPort())
		stencil.ThresholdBetween(0.5, 1.5)
		stencil.Update()

		nbins = 255
		binwidth = 0.01
		binstart = 0

		logging.debug ("image accum")
		accum = vtk.vtkImageAccumulate()
		accum.SetInputConnection(self.last_filter.GetOutputPort())
		accum.SetComponentSpacing(binwidth, binwidth, binwidth)
		accum.SetComponentExtent(0, nbins-1, 0, 0, 0, 0)
		accum.SetComponentOrigin(binstart, 0, 0)
		accum.SetIgnoreZero(True)
		#accum.SetStencilData(stencil.GetOutput())
		logging.debug ("image accum UPDATE")
		accum.Update()
		
		logging.debug ("to numpy")
		counts = vtk_to_numpy(accum.GetOutput().GetPointData().GetScalars())
		bins = np.linspace(binstart, (nbins-1)*binwidth, nbins)
		total = np.sum(counts)

		fxv = counts / total
		fx = counts / (total * binwidth)
		int_fx = np.zeros(fx.shape)
		for i in range(fx.shape[0]):
			int_fx[i] = np.sum(fx[0:i] * binwidth)            

		data = pd.DataFrame({
			"counts": counts,
			"fxv": fxv,
			"fx": fx,
			"integral_fx": int_fx
		}, index=bins)
		return data
		

class ParticlesPipeline:
	def __init__(self, fn="", time=0.0):		
		reader = PvdReader(fn)
		pvd_data = reader.get_data(time)

		glyphSrc = vtk.vtkSphereSource()		
		glyphSrc.SetRadius(0.002)

		glyph = vtk.vtkGlyph3D()
		glyph.SetSourceConnection(glyphSrc.GetOutputPort())
		glyph.SetInputDataObject(pvd_data)
		glyph.ScalingOff()
		last = glyph

		if pvd_data.GetClassName() == "vtkMultiBlockDataSet":

			geoFilter = vtk.vtkCompositeDataGeometryFilter()
			geoFilter.SetInputDataObject(pvd_data)
			geoFilter.Update()
			last = geoFilter	

		ctf = vtk.vtkColorTransferFunction()
		ctf.SetColorSpaceToDiverging()

		colorx = np.linspace(0, 1, 256)
		colorrgb = cm.get_cmap("viridis", len(colorx))
		for x in colorx:
			r,g,b,a = colorrgb(x)
			ctf.AddRGBPoint(x, r, g, b)

		mapper = vtk.vtkDataSetMapper()
		mapper.SetInputConnection(last.GetOutputPort())
		#mapper.SelectColorArray(variable_name)
		#mapper.SetColorModeToMapScalars()
		mapper.ScalarVisibilityOff()
		#mapper.SetLookupTable(ctf)
		#mapper.UseLookupTableScalarRangeOn()
		#mapper.SetScalarModeToUsePointFieldData()
		#mapper.InterpolateScalarsBeforeMappingOn()
		#mapper.SetScalarModeToUseCellFieldData()

		actor = vtk.vtkActor()
		actor.SetMapper(mapper)
		actor.GetProperty().SetRepresentationToSurface()
		actor.GetProperty().SetColor(1, 0, 0)

		self.reader = reader
		self.mapper = mapper
		self.actor = actor

class StaticStlPipeline:
	def __init__(self, fn="", opacity=0.5):
		reader = vtk.vtkSTLReader()
		reader.SetFileName(fn)
		reader.Update()
		mapper = vtk.vtkPolyDataMapper()
		mapper.SetInputConnection(reader.GetOutputPort())
		actor = vtk.vtkActor()
		actor.SetMapper(mapper)
		actor.GetProperty().SetRepresentationToSurface()
		actor.GetProperty().SetOpacity(opacity)

		self.reader = reader
		self.mapper = mapper
		self.actor = actor 

class MovingBodyPipeline:
	def __init__(self, fn="", time=0.0):

		reader = PvdReader(fn)
		pvd_data = reader.get_data(time)

		last = None

		if pvd_data.GetClassName() == "vtkMultiBlockDataSet":

			geoFilter = vtk.vtkCompositeDataGeometryFilter()
			geoFilter.SetInputDataObject(pvd_data)
			geoFilter.Update()
			last = geoFilter						

		ctf = vtk.vtkColorTransferFunction()
		ctf.SetColorSpaceToDiverging()

		colorx = np.linspace(0, 1, 256)
		colorrgb = cm.get_cmap("viridis", len(colorx))
		for x in colorx:
			r,g,b,a = colorrgb(x)
			ctf.AddRGBPoint(x, r, g, b)

		mapper = vtk.vtkPolyDataMapper()
		if last is None:
			mapper.SetInputDataObject(pvd_data)
		else:
			mapper.SetInputConnection(last.GetOutputPort())

		#mapper.SelectColorArray(variable_name)
		#mapper.SetColorModeToMapScalars()
		#mapper.ScalarVisibilityOn()
		#mapper.SetLookupTable(ctf)
		#mapper.UseLookupTableScalarRangeOn()
		#mapper.SetScalarModeToUsePointFieldData()
		#mapper.InterpolateScalarsBeforeMappingOn()
		#mapper.SetScalarModeToUseCellFieldData()

		actor = vtk.vtkActor()
		actor.SetMapper(mapper)
		actor.GetProperty().SetRepresentationToSurface()

		self.reader = reader
		self.mapper = mapper
		self.actor = actor

class BlockDataPipeline:

	def __init__(self, fn="", time=0.0, doResample=True, color=None):

		reader = PvdReader(fn)
		pvd_data = reader.get_data(time)

		if pvd_data.GetClassName() == "vtkMultiBlockDataSet":

			geoFilter = vtk.vtkCompositeDataGeometryFilter()
			geoFilter.SetInputDataObject(pvd_data)
			geoFilter.Update()
			last = geoFilter

			if doResample:
				nx, ny, nz = GetNiceSamplingBounds(pvd_data)				
				resample = vtk.vtkResampleToImage()
				resample.UseInputBoundsOn()
				resample.SetSamplingDimensions(nx, ny, nz)
				resample.SetInputConnection(last.GetOutputPort())
				resample.Update()
				last = resample

		else:

			c2p = vtk.vtkCellDataToPointData()
			c2p.SetInputDataObject(pvd_data)
			last = c2p

		thr = vtk.vtkThreshold()			
		thr.SetInputConnection(last.GetOutputPort())
		thr.ThresholdByUpper(0.5)
		thr.SetInputArrayToProcess(0, 0, 0, vtk.vtkDataObject.FIELD_ASSOCIATION_POINTS, "Volume Fraction (-)")
		last = thr

		if color is None:
			color = vtk.vtkColorTransferFunction()
			color.SetColorSpaceToDiverging()

			colorx = np.linspace(0, 1, 256)
			colorrgb = cm.get_cmap("viridis", len(colorx))
			for x in colorx:
				r,g,b,a = colorrgb(x)
				color.AddRGBPoint(x, r, g, b)

		mapper = vtk.vtkDataSetMapper()
		mapper.SetInputConnection(last.GetOutputPort())
		#mapper.SelectColorArray(variable_name)
		#mapper.SetColorModeToMapScalars()
		mapper.ScalarVisibilityOn()
		mapper.SetLookupTable(color)
		#mapper.UseLookupTableScalarRangeOn()
		mapper.SetScalarModeToUsePointFieldData()
		mapper.InterpolateScalarsBeforeMappingOn()
		#mapper.SetScalarModeToUseCellFieldData()

		actor = vtk.vtkActor()
		actor.SetMapper(mapper)
		actor.GetProperty().SetRepresentationToSurface()

		self.reader = reader
		self.mapper = mapper
		self.actor = actor

class MStarResults:
	def __init__(self, outdir="out", timegroup="Slice"):
		self.outdir = outdir
		self.output = os.path.join(outdir, "Output")
		self.stats = os.path.join(outdir, "Stats")
		self.timegroup = timegroup
		self.reload()
		self.ctf = {}
		self.opacity = {}
		self.variable = ""
		self.textprop = None
		self.show_statics = True
		self.time = 0

	def set_show_statics(self, val):
		self.show_statics = val

	def get_color(self):
		return self.get_scalar_bar(self.variable)
	
	def get_opacity(self):
		if self.variable in self.opacity:
			return self.opacity[self.variable]
		return 1.0

	def set_opacity(self, varname, xs, ys):
		opfunc = vtk.vtkPiecewiseFunction()
		for x,y in zip(xs, ys):
			opfunc.AddPoint(x, y)
		self.opacity[varname] = opfunc

	def set_variable(self, varname=""):
		self.variable = varname
		
	def set_ctf(self, varname="", cmap=None, mplnorm=None):
		ctf = vtk.vtkColorTransferFunction()		
				
		sampled = np.linspace(mplnorm.vmin, mplnorm.vmax, cmap.N)		
		for i in range(cmap.N):			
			r,g,b,a = cmap(mplnorm(sampled[i]))
			ctf.AddRGBPoint(sampled[i], r, g, b)

		self.ctf[varname] = ctf

	def set_scalar_bar(self, varname="", minlimit=0, maxlimit=1, colorscale="viridis"):
		ctf = vtk.vtkColorTransferFunction()		
		
		ncolors = 256
		sampled = np.linspace(minlimit, maxlimit, ncolors)
		norm = mpl.colors.Normalize(minlimit, maxlimit)		
		c = cm.get_cmap(colorscale, ncolors)
		for i in range(ncolors):			
			r,g,b,a = c(norm(sampled[i]))
			ctf.AddRGBPoint(sampled[i], r, g, b)

		self.ctf[varname] = ctf

	def get_scalar_bar(self, varname=""):
		if len(varname) == 0:
			return self.ctf[self.variable]

		if varname in self.ctf:
			return self.ctf[varname]
		return vtk.vtkColorTransferFunction()

	def reload(self):
		self.files = []
		for fn in glob.glob(os.path.join(self.output, "*.*")):
			if os.path.isfile(fn):				
				m = MStarFile(fn)
				logging.debug( str( m ))
				self.files.append(m)
		
	def print_summary(self):
		for f in self.files:
			print(f)
		
	def get_group_times(self, grpname=""):
		for p in self.files:
			if p.timegroup == grpname:
				r = PvdReader(p.filename)
				return r.get_times()
		return []
	
	def get_times(self):
		return self.get_group_times(self.timegroup)	

	def set_time(self, time):
		self.time = time

	def get_time(self):
		return self.time

	def get_dt(self):
		t = self.get_times()
		if len(t) >= 2:
			return t[1] - t[0]
		return 0

	def set_variable(self, varname="", vartype='POINT'):
		self.variable = varname
		self.variable_type = vartype

	def set_text_property(self, prop):
		self.textprop = prop

	def get_variable_ranges(self):
		# Get variable ranges
		# yields tuples of (variablename, min, max)

		vars = {}
		for f in self.files:
			if f.type in [ "X", "Y", "Z", "Volume" ]:
				r = f.data_ranges(self.time)
				if r:
					for ctype,cname,comp,minv,maxv in r:
						key = (cname, comp, ctype)
						#key = cname
						compmin = minv
						compmax = maxv

						if key in vars:
							compmin = min(vars[key][0], compmin)
							compmax = max(vars[key][1], compmax)
												
						vars[key] = (compmin, compmax)

		for k,v in vars.items():	
			#yield (k, v[0], v[1])
			yield (*k, *v)


	def add_scalar_bar(self, renderer):
		scalarBar = vtk.vtkScalarBarActor()
		scalarBar.SetLookupTable(self.ctf[self.variable])		
		scalarBar.SetNumberOfLabels(5)		

		if self.textprop is not None:
			scalarBar.AnnotationTextScalingOff()
			scalarBar.SetLabelTextProperty(self.textprop)
			scalarBar.SetAnnotationTextProperty(self.textprop)
			scalarBar.SetTitleTextProperty(self.textprop)

		renderer.AddActor2D(scalarBar)

	def plot(self, renderer):

		grps = [ self.timegroup ]
		if self.show_statics:
			grps.append("Static")

		for f in self.files:

			if f.timegroup in grps:
				f.add_pipeline(self, renderer)				

	def get_volume_pipeline(self):
		for f in self.files:
			if f.timegroup in [ self.timegroup ] and f.type == "Volume":
				return f.get_pipeline(self)

	def get_slice_pipelines(self, time):
		
		for f in self.files:
			if f.timegroup in  [ "Slice", "Static" ] :
				yield f.make_pipeline(time)


def mymax(x,y):
	if x > y:
		return x
	return y


def reset_camera_tight2(renderer, margin_factor=1.02, bounds=None):
	""" Resets camera so the content fit tightly within the window.

	Parameters
	----------
	margin_factor : float (optional)
		Margin added around the content. Default: 1.02.

	"""
	renderer.ComputeAspect()
	cam = renderer.GetActiveCamera()
	aspect = renderer.GetAspect()

	X1, X2, Y1, Y2, Z1, Z2 = renderer.ComputeVisiblePropBounds()

	if bounds is None:
		X1, X2, Y1, Y2, Z1, Z2   = renderer.ComputeVisiblePropBounds()
	else:
		X1, X2, Y1, Y2, Z1, Z2  = bounds

	width, height = X2-X1, Y2-Y1
	center = np.array((X1 + width/2., Y1 + height/2., 0))

	angle = np.pi*cam.GetViewAngle()/180.
	dist = max(width/aspect[0], height) / np.sin(angle/2.) / 2.
	position = center + np.array((0, 0, dist*margin_factor))

	cam.SetViewUp(0, 1, 0)
	cam.SetPosition(*position)
	cam.SetFocalPoint(*center)
	renderer.ResetCameraClippingRange(X1, X2, Y1, Y2, Z1, Z2)

	parallelScale = max(width/aspect[0], height) / 2.
	cam.SetParallelScale(parallelScale*margin_factor)

def create_camera_tight(eye_index=0, eye_flip=1, up_index=1, bounds=None):
	""" 
	Parameters
	----------
	eye_index: 0, 1, 2  for x y or z
	eye_flip: set negative for flip direction
	up_index: 0, 1 2  which way is up?

	"""	

	camera = vtk.vtkCamera()
	padding_factor = 1.1

	directions = [0, 1, 2]
	directions.remove(eye_index)
	directions.remove(up_index)
	width_index = directions[0]

	if bounds is None:
		raise ValueError("bounds must be provided")
	else:
		xmn,xmx,ymn,ymx,zmn,zmx = bounds

	xl = xmx - xmn
	yl = ymx - ymn
	zl = zmx - zmn
	l = [xl, yl, zl]
	xc = xmn + xl * 0.5
	yc = ymn + yl * 0.5
	zc = zmn + zl * 0.5

	tanangle = math.tan(camera.GetViewAngle() * math.pi / 180.0 / 2.0)
	xeye = xl * padding_factor / (2 * tanangle)
	yeye = yl * padding_factor / (2 * tanangle)
	zeye = zl * padding_factor / (2 * tanangle)
	eye_d = [ xeye, yeye, zeye ]

	aspect_ratio = 1.0 
	needed_height = aspect_ratio * l[width_index]

	eye = [ xc, yc, zc ]
	up = [0,0,0]
	height = mymax(needed_height, l[up_index]) * 0.5
	up[up_index] = 1
	eye[eye_index] += max(eye_d) * eye_flip	

	camera.SetParallelProjection(True)	
	camera.SetParallelScale(height)		
	camera.SetViewUp(up)
	camera.SetPosition(eye)
	camera.SetFocalPoint(xc,yc,zc)

	return camera

def reset_camera_tight(renderer, eye_index=0, eye_flip=1, up_index=1, bounds=None):
	""" Resets camera so the content fit tightly within the window.

	Parameters
	----------
	eye_index: 0, 1, 2  for x y or z
	eye_flip: set negative for flip direction
	up_index: 0, 1 2  which way is up?

	"""	
	camera = renderer.GetActiveCamera()
	aspect = renderer.GetAspect()	
	padding_factor = 1.1

	directions = [0, 1, 2]
	directions.remove(eye_index)
	directions.remove(up_index)
	width_index = directions[0]

	if bounds is None:
		xmn,xmx,ymn,ymx,zmn,zmx  = renderer.ComputeVisiblePropBounds()
	else:
		xmn,xmx,ymn,ymx,zmn,zmx = bounds

	xl = xmx - xmn
	yl = ymx - ymn
	zl = zmx - zmn
	l = [xl, yl, zl]
	#print (xmn,xmx,ymn,ymx,zmn,zmx)
	#print (l)

	xc = xmn + xl * 0.5
	yc = ymn + yl * 0.5
	zc = zmn + zl * 0.5

	tanangle = math.tan(camera.GetViewAngle() * math.pi / 180.0 / 2.0)
	xeye = xl * padding_factor / (2 * tanangle)
	yeye = yl * padding_factor / (2 * tanangle)
	zeye = zl * padding_factor / (2 * tanangle)
	eye_d = [ xeye, yeye, zeye ]

	aspect_ratio = 1.0 #float(view.ViewSize[1]) / float(view.ViewSize[0])
	needed_height = aspect_ratio * l[width_index]

	eye = [ xc, yc, zc ]
	up = [0,0,0]
	#print (needed_height, l, up_index, l[up_index])
	height = mymax(needed_height, l[up_index]) * 0.5

	up[up_index] = 1
	eye[eye_index] += max(eye_d) * eye_flip

	#print ("-- Updating camera: ", "height=", height, "up=", up, "pos=", eye, "focal=", (xc,yc,zc))

	camera.SetParallelProjection(True)	
	camera.SetParallelScale(height)		
	camera.SetViewUp(up)
	camera.SetPosition(eye)
	camera.SetFocalPoint(xc,yc,zc)

def dumpActor2d(a):
	print("GetPosition()=", a.GetPosition())
	print("GetPosition2()=", a.GetPosition2())
	print("H=", a.GetHeight(), "W=", a.GetWidth())

def create_scalarbar_image(ctf, size=(100,100), textprop=None):
	colors = vtk.vtkNamedColors()

	scalarBar = vtk.vtkScalarBarActor()
	scalarBar.SetLookupTable(ctf)		
	scalarBar.SetNumberOfLabels(5)	
	scalarBar.SetTitleTextProperty(textprop)
	scalarBar.SetLabelTextProperty(textprop)
	scalarBar.SetAnnotationTextProperty(textprop)	
	scalarBar.SetPosition(0, 0.01)
	scalarBar.SetWidth(0.95)
	scalarBar.SetHeight(0.95)

	renderer = vtk.vtkRenderer()	
	renderer.SetBackground(colors.GetColor3d("White"))    	
	renderer.AddActor2D(scalarBar)	

	renderWindow = vtk.vtkRenderWindow()
	renderWindow.AddRenderer(renderer)
	renderWindow.SetOffScreenRendering(1)
	renderWindow.SetSize( int(size[0]/2), int(size[1]/2) )
	renderWindow.SetMultiSamples(4)
	renderWindow.Render()

	windowto_image_filter = vtk.vtkWindowToImageFilter()
	windowto_image_filter.SetInput(renderWindow)
	windowto_image_filter.SetScale(2)	

	writer = vtk.vtkPNGWriter()
	writer.SetInputConnection(windowto_image_filter.GetOutputPort())
	writer.WriteToMemoryOn()
	writer.Write()

	buf = io.BytesIO()
	buf.write(writer.GetResult())
	buf.seek(0)

	return buf

def f2icolor(x=1.0):
	return int( x * 255 )

def color2int(rgb=(0.0,0.0,0.0)):
	return ( f2icolor( rgb[0]), f2icolor( rgb[1] ), f2icolor( rgb[2] ))

def draw_color_bar(ctf=None, opacityFunc=None, size=(10,100), nticks=5, min=0.0, max=1.0, fontsize=16):
	img = Image.new("RGBA", size=size)	

	font = ImageFont.truetype("arial.ttf", fontsize)
	dr = ImageDraw.Draw(img)

	barwidth = 0.3333

	w = size[0] * barwidth
	n = size[1]
	for i in range(size[1]):
		x = min + i * (max - min) / n
		rgbflt = [ 0.0, 0.0, 0.0 ]
		ctf.GetColor(x, rgbflt)
		rgb = color2int(rgbflt)
		opacity =  f2icolor ( 1.0 )

		if opacityFunc is not None:
			opacity = f2icolor (  opacityFunc.GetValue(x) )

		rgba = (rgb[0], rgb[1], rgb[2], opacity)
		dr.line([ (0, n-i), (w, n-i )  ], fill=rgba )

	anchors = [ "lm" ] * nticks
	anchors[0] = "ls"
	anchors[-1] = "lt"
	values = np.linspace(min, max, nticks)

	for i in range(nticks):
		y = n - i * (n / (nticks - 1))
		val = "{0:.3g}".format( values[i] )
		dr.text(xy=(w+1, y), text=val, anchor=anchors[i], font=font, fill="black")

	#dr.text( (0,h), title , anchor='la', font=font )
	return img

def make_frames(caseDir="", timegroup="", variables=[]):
	"""
	"""

	results = MStarResults(timegroup=timegroup, outdir=os.path.join(caseDir, "out"))

	title = os.path.basename(caseDir)

	#  Output directory to place all data and images
	output_dir = os.path.join(caseDir, "out/Processed")

	# Define variable names, min, max contour values
	#variables = variables

	# Get all variables, using auto-computed ranges based on last available time step
	#variables = results.get_variable_ranges()

	# (name, eye direction, up direction, flip)
	views = [
		("front", 0, 1, False),
		#("back", 0, 1, True),
		("top", 1, 2, False),
		#("bottom", 1, 2, True),
		#("side_1", 2, 1, False),
		("side_2", 2, 1, True),
	]

	img_size = (500,500)

	colors = vtk.vtkNamedColors()

	textprop = vtk.vtkTextProperty()
	textprop.SetFontSize(48)
	textprop.SetFontFamilyToArial()
	textprop.ItalicOff()
	textprop.ShadowOff()
	textprop.BoldOff()
	textprop.SetLineSpacing(1.1)
	textprop.SetColor(colors.GetColor3d("Black"))

	results.set_text_property(textprop)

	for variable_name,vmin,vmax,cmapname in variables:

		var_name = NiceVariableName(variable_name)
		var_out_dir = os.path.join(output_dir, var_name)	
		os.makedirs(var_out_dir, exist_ok=True)
					
		results.set_scalar_bar(variable_name, vmin, vmax, cmapname)
		results.set_opacity(variable_name, [vmin, vmax], [0, 1.0])			
		results.set_variable(variable_name)		
		results.set_show_statics(False)

		#scalar_bar_img = Image.open( create_scalarbar_image(results.get_scalar_bar(), (50,img_size[1]) , textprop))
		scalar_bar_img_2 = draw_color_bar(results.get_scalar_bar(), opacityFunc=results.get_opacity(), size=(75, img_size[1]), min=vmin, max=vmax)	

		# Render the frames directly into a video file
		img_index = 0
		
		for t in results.get_times():
			
			camera = vtk.vtkCamera()
			camera.SetPosition(1, 1, 1)
			camera.SetFocalPoint(0, 0, 0)

			renderer = vtk.vtkRenderer()
			text_annotation = "  ".join( ( title, variable_name, "t = {0}".format(t) ) )
			#textMapperL.SetInput("\n".join( ( title, variable_name, "t={0}".format(t) ) ))
			renderer.SetBackground(colors.GetColor3d("White"))    
			#renderer.AddActor2D(textActorL)			
			renderer.SetActiveCamera(camera)
			
			results.set_time(t)
			results.plot(renderer)		

			axes = vtk.vtkAxesActor()

			renderWindow = vtk.vtkRenderWindow()
			renderWindow.AddRenderer(renderer)
			renderWindow.SetOffScreenRendering(1)
			renderWindow.SetSize( int(img_size[0]/2), int(img_size[1]/2) )
			renderWindow.SetMultiSamples(4)

			# An interactor
			renderWindowInteractor = vtk.vtkRenderWindowInteractor()
			renderWindowInteractor.SetRenderWindow(renderWindow)

			# lower left axis orientation
			widget = vtk.vtkOrientationMarkerWidget()
			rgba = [0] * 4
			colors.GetColor('Carrot', rgba)
			widget.SetOutlineColor(rgba[0], rgba[1], rgba[2])
			widget.SetOrientationMarker(axes)
			widget.SetInteractor(renderWindowInteractor)
			widget.SetViewport(0.0, 0.0, 0.1, 0.1)
			widget.SetEnabled(1)			
			
			for v in views:
				view_name = v[0]
				view_eye = v[1]
				view_up = v[2]
				view_flip = v[3]

				renderWindow.Render()

				windowto_image_filter = vtk.vtkWindowToImageFilter()
				windowto_image_filter.SetInput(renderWindow)
				windowto_image_filter.SetScale(2)

				reset_camera_tight(renderer, eye_index=view_eye, up_index=view_up)

				writer = vtk.vtkPNGWriter()
				writer.SetInputConnection(windowto_image_filter.GetOutputPort())
				writer.WriteToMemoryOn()
				writer.Write()

				buf = io.BytesIO()
				buf.write(writer.GetResult())
				buf.seek(0)

				# Read data into PIL image buffer
				vtk_img = Image.open(buf)
				
				w = vtk_img.size[0]
				h = vtk_img.size[1]
				
				total_w = w + scalar_bar_img_2.size[0]
				total_h = h + 20

				# Stack VTK image with scalar bar and annotation in a new image		
				com_img = Image.new("RGBA", size=(total_w, total_h), color=(255,255,255))
				com_img.paste(vtk_img, (0,0) )
				com_img.alpha_composite(scalar_bar_img_2.copy(), dest=(w,0))

				font = ImageFont.truetype("arial.ttf", 15)
				dr = ImageDraw.Draw(com_img)

				dr.text( (0,h), title , anchor='la', font=font, fill="black" )
				dr.text( (total_w/2,h), variable_name , anchor='ma', font=font, fill="black" )
				dr.text( (total_w,h), "t = {0}".format(t) , anchor='ra', font=font , fill="black")
							
				fn = "{name}_{view_name}_{n:05}.png".format(name=var_name, n=img_index, view_name=view_name)
				fn = os.path.join(var_out_dir, fn)		
				print ("-- Rendering time: ", t, "-->", fn)
				com_img.save(fn)

			img_index += 1
		
		
		
		
	





from typing import Optional
from pydantic import BaseModel, Field
from pydantic.dataclasses import dataclass
from enum import Enum, IntEnum


def ToFloatVec(str="", delim=","):
	if str is None:
		return tuple()
	return tuple(map(float, str.split(delim)))

def ToIntVec(str="", delim=","):
	if str is None:
		return tuple()
	return tuple(map(int, str.split(delim)))

def ToColorVec(str="", delim=","):
	return [ max(0.0, min(1.0, v / 255.0)) for v in ToFloatVec(str, delim) ]

def ToBool(v):
	if isinstance(v, bool):
		return v
	elif isinstance(v, str):
		return v.lower() in ("yes", "true", "t", "1")
	elif isinstance(v, int):
		return v != 0
	
	raise ValueError("Cannot parse value to boolean: " + str(v))

def smootherStep(e0, e1, x):
	x = min(1.0, max(0.0, (x - e0) / (e1 - e0)))
	return x * x * x * (x * (x * 6 - 15) + 10)

class VTKVariable(BaseModel):
	name: str
	min: float
	max: float
	number_components: Optional[int]
	component: int = 0
	component_name: Optional[str]
	display_name: str = ""
	color_scale: str = "magma"
	discrete: bool = False
	discrete_n: int = 12

	def get_component_name(self):
		if self.component_name is not None:
			return self.component_name
		elif self.number_components is not None:
			if self.number_components == 3:
				if self.component < 3:
					return ["X", "Y", "Z"][self.component]
			elif self.number_components == 1:
				return ""
			else:
				return str(self.component)
		return str(self.component)
				

def ColorInt2Float(v):
	return max(0.0, min(1.0, v / 255.0))

class RGBIntValue(BaseModel):
	red: int = Field(0, ge=0, le=255)
	green: int = Field(0, ge=0, le=255)
	blue: int = Field(0, ge=0, le=255)

	def ToFloatValue(self) -> tuple[float, float, float]:
		return (ColorInt2Float(self.red), ColorInt2Float(self.green), ColorInt2Float(self.blue))

class OpacityFuncType(str, Enum):
	smooth = 'smooth'
	ramp = 'ramp'
	constant = 'constant'

class DirEnum(str, Enum):
	xn = 'xn'
	xp = 'xp'
	yn = 'yn'
	yp = 'yp'
	zn = 'zn'
	zp = 'zp'
		
	def direction(self):		
			m = { DirEnum.xn:0, 
					DirEnum.xp:0, 
					DirEnum.yn:1, 
					DirEnum.yp:1, 
					DirEnum.zn:2, 
					DirEnum.zp:2 }
			return m[self]
		
	def sign(self):		
		m = { DirEnum.xn:-1, 
				DirEnum.xp:1, 
				DirEnum.yn:-1, 
				DirEnum.yp:1, 
				DirEnum.zn:-1, 
				DirEnum.zp:1 }
		return m[self]

class TimesEnum(str, Enum):
	all = 'all'
	last = 'last'

class ColorByEnum(str, Enum):
	variable = "variable"
	solid_color = "solid_color"

class OpacityFunction(BaseModel):
	type: OpacityFuncType = OpacityFuncType.smooth
	start: float = Field(0.0, ge=0.0, le=1.0)
	end: float = Field(1.0, ge=0.0, le=1.0)
	constant_value: float = Field(1.0, ge=0.0, le=1.0)

class BlockDataPipeline(BaseModel):
	file: str
	color_by: ColorByEnum

class MovingBodyPipeline(BaseModel):
	file: str
	color_by: ColorByEnum

class ParticlesPipeline(BaseModel):
	file: str
	color_by: ColorByEnum

class StaticStlPipeline(BaseModel):
	file: str
	solid_color: RGBIntValue
	solid_opacity: float = Field(..., ge=0.0, le=1.0)

class VolumePipeline(BaseModel):
	file: str	
	opacity_function: OpacityFunction = OpacityFunction()

class Watermark(BaseModel):
	filename: str
	location: str = "ul"
	alpha: float = 0.25


class ViewStandard(BaseModel):	
	name: str
	eye_side: DirEnum
	up: DirEnum	


class ViewCustom(BaseModel):
	name: str
	eye_position: tuple[float, float, float]
	look_at: tuple[float, float, float]
	up: tuple[float, float, float]
	eye_position: tuple[float, float, float]
	fit: bool = False

class VideoOption(BaseModel):
	enabled: bool = False
	speed: float = Field(1.0, gt=0.0)
	quality: int = Field(20, gt=1, lt=30)

class VTKPlotConfig(BaseModel):

	name: str
	size_x: int = Field(1024, ge=10)
	size_y: int = Field(1024, ge=10)
	image_output: bool = True
	time: TimesEnum = TimesEnum.all
	background_color: RGBIntValue = RGBIntValue()

	variables: list[VTKVariable] = []
	block_pipelines: list[BlockDataPipeline] = list()
	moving_pipelines: list[MovingBodyPipeline]=  list()
	particle_pipelines: list[ParticlesPipeline]=  list()
	stl_pipelines: list[StaticStlPipeline]=  list()
	volume_pipelines: list[VolumePipeline]=  list()
	
	standard_views: list[ViewStandard] = []
	custom_views: list[ViewCustom] = []

	watermarks: list[Watermark] = []
	video_options: VideoOption = VideoOption()

class PlotAxisEnum(str, Enum):
	linear = "linear"
	log10 = "log10"


class StatPlotDataSeries(BaseModel):

	filename: str
	x: Optional[str]
	y: Optional[str]
	x_column: Optional[int] = 0
	y_column: Optional[int] = 0
	line_color: Optional[RGBIntValue]
	line_style: Optional[str] = "solid"
	line_width: Optional[float]
	name: Optional[str]

class LegendLocEnum(str, Enum):
	best = "best"
	upper_left = "upper left"
	upper_right = "upper right"
	lower_left = "lower left"
	lower_right = "lower right"
	upper_center = "upper center"
	lower_center = "lower center"
	center_left = "center left"
	center_right = "center right"
	center = "center"

class StatPlotStyle(BaseModel):
	name: Optional[str]
	size: Optional[tuple[float, float]]
	style: Optional[list[str]]
	legend_location: Optional[LegendLocEnum]

	def override(self, rhs):		
		if rhs is None:
			return self
		s = self.copy()
		if rhs.size is not None:
			s.size = rhs.size
		if rhs.style is not None:
			s.style = rhs.style
		if rhs.legend_location is not None:
			s.legend_location = rhs.legend_location
		return s


class StatPlot(BaseModel):

	name: str
	y_title: Optional[str]
	x_title: Optional[str]
	chart_title: Optional[str]
	x_lim: Optional[tuple[float, float]]
	y_lim: Optional[tuple[float, float]]	
	y_axis: PlotAxisEnum = PlotAxisEnum.linear
	series: list[StatPlotDataSeries]
	style: Optional[StatPlotStyle]


class ResultStatImage(BaseModel):

	name: str
	config: StatPlot
	image_filename: str

class ResultAutoStatImage(BaseModel):

	name: str
	stat_filename: str
	image_filename: str

class ResultVtkImage(BaseModel):
	time: float
	image_filename: str
	config: VTKPlotConfig
	view: str	

class ResultsProcessed(BaseModel):
	custom_stat_plots: list[ResultStatImage] = []
	auto_stat_plots: list[ResultAutoStatImage] = []
	vtk_images: list[ResultVtkImage] = []

class BatchPostConfig(BaseModel):
	plots: list[VTKPlotConfig] = []		
	default_stat_plot_style: Optional[StatPlotStyle]
	stat_plots: list[StatPlot] = []
	auto_stat_plots: bool = False
	auto_pdf_report: bool = False

class AutoReportConfig(BaseModel):
	title: str
	results: ResultsProcessed	
	model_tree_filename: Optional[str]





from datetime import datetime
import os
from itertools import groupby
import xml.etree.ElementTree as ET

from reportlab.platypus import SimpleDocTemplate, Paragraph, Spacer, PageBreak
from reportlab.platypus.doctemplate import PageTemplate, BaseDocTemplate
from reportlab.platypus.tableofcontents import TableOfContents
from reportlab.platypus.frames import Frame
from reportlab.platypus.tables import Table, TableStyle
from reportlab.platypus.figures import Figure
from reportlab.platypus.flowables import Image
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.lib import colors
from reportlab.rl_config import defaultPageSize
from reportlab.lib.units import inch, cm
from reportlab.graphics.charts.lineplots import LinePlot,ScatterPlot
from reportlab.graphics.widgets.markers import makeMarker
from reportlab.graphics.shapes import Drawing   

from .config import AutoReportConfig
from .config import ResultAutoStatImage
	

class MyDocTemplate(BaseDocTemplate):
	def __init__(self, filename, **kw):
		self.allowSplitting = 0
		BaseDocTemplate.__init__(self, filename, **kw)
		template = PageTemplate('normal', [Frame(1*inch, 1*inch, 6*inch, 9*inch, id='F1')])
		self.addPageTemplates(template)
		self.current_key_index = 1
	
	def afterFlowable(self, flowable):
		if flowable.__class__.__name__ == "Paragraph":
			text = flowable.getPlainText()
			style = flowable.style.name
			
			if style == "Heading1":
				self.current_key_index += 1
				key = str(self.current_key_index)
				self.canv.bookmarkPage(key)
				self.canv.addOutlineEntry(text, key, 0, 0)
				self.notify("TOCEntry", (0, text, self.page))
			elif style == "Heading2":
				self.current_key_index += 1
				key = str(self.current_key_index)
				self.canv.bookmarkPage(key)
				self.canv.addOutlineEntry(text, key, 1, 0)
				self.notify("TOCEntry", (1, text, self.page))
			elif style == "Heading3":
				self.current_key_index += 1
				key = str(self.current_key_index)
				self.canv.bookmarkPage(key)
				self.canv.addOutlineEntry(text, key, 2, 0)
				self.notify("TOCEntry", (2, text, self.page))

def get_model_tree(modelFn = ""):
	SKIPPED_CATS = [ "Object View", "Geometry View", "View" , "Display Attributes" ]
	SKIPPED_NAMES = [ "Name", "Enabled" ]
	tree = ET.parse(modelFn)
	root = tree.getroot()
	model = root.find("model")
	
	def sort_by_preferred(x):
		preferred_ordering = [            
			"LatticeDomain",
			"VoxelizedBody",
			"ImmersedBody",
			"SimulationParameters",
			"ParticleZone",
			"BubbleZone",
			"ScalarField",
		]
		if x in preferred_ordering:
			return preferred_ordering.index(x)
		else:
			return len(preferred_ordering)       
	
	def _walk_component(c, component_tables, level):        
		level += 1
		table = [ ["Properties", "Value", "Units" ]  ]        
		
		cat_props = {}

		for p in c.findall("./property"):
			name = p.get("name", "")
			cat = p.get("category", "General")
			units = p.get("units", "")                       
			
			if any([ c in cat for c in SKIPPED_CATS]):
				continue             
			if any([ n in name for n in SKIPPED_NAMES]):
				continue
				
			value = ""
			if p.text is not None:
				value = p.text
				
			table.append([ name, value,  units ])

			if cat not in cat_props:
				cat_props[cat] = []

			cat_props[cat].append([name, value, units])
		
		component_tables.append({
			'name':c.get("name"),
			'table': table,
			'categorized_tables': cat_props,
			'id': "component_"
		})
				
		for child in c.findall("./component"):
			_walk_component(child, component_tables, level)
		
	component_tables = []    
	for c in model.findall("./component"):
		_walk_component(c, component_tables, 1)  
	
	screen_shots = []
	out_dir = os.path.dirname(modelFn)
	for img in root.findall("./screenShots/image"):
		fname = os.path.join(out_dir, img.get("filename"))
		screen_shots.append(fname)

	return component_tables,screen_shots

def create_auto_report(conf: AutoReportConfig, outputfn="Report.pdf"):
	# Report style variables
	styles = getSampleStyleSheet()
	style = styles["Normal"]
	h1 = styles["Heading1"]
	h2 = styles["Heading2"]
	h3 = styles["Heading3"]
	h4 = styles["Heading4"]
	h5 = styles["Heading5"]
	normal = styles["Normal"]

	report = []
	report.append( Spacer(1, 2*inch) )
	report.append( Paragraph(conf.title, h1) )
	report.append( Spacer(1, 2*inch) )
	report.append( Paragraph( "Created: " + datetime.now().strftime("%x - %X"), normal) )
	report.append( Paragraph( conf.title, normal) )

	msbname = ""
	version = ""
	if conf.model_tree_filename is not None and os.path.isfile(conf.model_tree_filename):
		reportxml = ET.parse(conf.model_tree_filename).getroot()
		if (n := reportxml.find("./model")) is not None:
			report.append( Paragraph( "M-Star CFD Version: " + n.get("version", ""), normal) )			
			report.append( Paragraph( "MSB File: " + n.get("filename", ""), normal) )			


	# Report parts
	
	report.append( PageBreak() )

	# Table of contents
	toc = TableOfContents()
	toc.levelStyles = [h1, h2]
	report.append(toc)
	report.append(PageBreak())

	# MSB summary: images, properties listing
	if conf.model_tree_filename is not None and os.path.isfile(conf.model_tree_filename):

		objTree,screens = get_model_tree(conf.model_tree_filename)		
		report.append(Paragraph("Model Summary", h1))

		for screenfn in screens:
			report.append(Image(screenfn, width=4*inch, height=4*inch, kind='bound'))

		for compProps in objTree:
			t = Table(compProps["table"])
			t.setStyle(TableStyle([("ROWBACKGROUNDS", (0,0), (-1,-1), (colors.HexColor(0xe6e6e6), colors.HexColor(0xffffff)))]))
			
			report.append(Paragraph(compProps["name"], h2))
			report.append(t)
		

	# Custom stat plots
	report.append(Paragraph("Stats Plots", h1))
	for plot in conf.results.custom_stat_plots:
		report.append(Paragraph(plot.name, h2))
		img = Image(plot.image_filename, width=6*inch, height=6*inch, kind='bound')
		report.append(img)        				

	# VTK images (only last time shown) #TODO
	#report.append(Paragraph("VTK Data", h1))

	# Auto stat plots
	def getStatFilename(g: ResultAutoStatImage):
		return g.stat_filename

	auto_stat_plots = sorted(conf.results.auto_stat_plots, key=getStatFilename)

	report.append(Paragraph("Auto Stats Plots", h1))
	for key,group in groupby(auto_stat_plots, key=getStatFilename):
		report.append(Paragraph(os.path.basename(key), h2))
		for plot in group:
			report.append(Paragraph(plot.name, h3))	
			img = Image(plot.image_filename,  width=6*inch, height=6*inch, kind='bound')
			report.append(img)   
	
	print ("-- Writing report PDF:", outputfn)
	doc = MyDocTemplate(outputfn)
	doc.multiBuild(report)
